package fr.doap.jdb;

import fr.doap.slog.*;

import java.sql.*;
import java.util.*;
import java.util.concurrent.*;
import java.util.function.*;
import java.util.stream.*;

public class DBWrap {
	public static String sJID = "_JID_";

	public DBWrap(String aConn, String aUser, String aPass) {
		try {
			if (mConnection != null) mConnection.close();
			mConnection = DriverManager.getConnection(aConn+";DB_CLOSE_ON_EXIT=FALSE", aUser, aPass);
			mExecutorService = Executors.newSingleThreadExecutor();
			
			Runtime.getRuntime().addShutdownHook(new Thread(() -> {
				mExecutorService.shutdown();
				L.t("Waiting for queue to finish...");
				try { while (!mExecutorService.awaitTermination(100, TimeUnit.MILLISECONDS)) ;	} catch (Exception aE) { L.t(aE); }
				L.t("Queue empty... finishing!");
				
				try {
					if ((mConnection != null)&&(!mConnection.isClosed())) {
						mConnection.close();
						L.d("Connection closed");
					}
				} catch (Exception aE) { L.w(aE); }
			}));
			
			ResultSet lRST = mConnection.getMetaData().getTables(null, null, "OBJECTS", null);
			if (!lRST.next()) exec("CREATE TABLE OBJECTS ("+sJID+" INT PRIMARY KEY)");
			
			exec(() -> {
				try {
					ResultSet lRSPK = mConnection.getMetaData().getColumns(null, null, "OBJECTS", null);
					while (lRSPK.next()) mColumns.add(lRSPK.getString(4));
				} catch (Exception aE) { L.w(aE); }
			});
//			System.out.println("Registered Primary keys: "+ sPrimaryKeys);
		
		} catch (Exception aE) {
			aE.printStackTrace();
		}
	}
	
	public DBWrap setSQLTypeResolver(Function<String, String> aFun) {
		mColumnSQLTypeResolver = aFun;
		return this;
	}
	
	public void exec(String aS) {
		mExecutorService.submit(() -> {
			try {
				mConnection.createStatement().execute(aS);
				L.d("SQL: "+aS);
			} catch(Exception aE) { L.w(aE); }});
	}
	
	public void exec(Runnable aR) {
		mExecutorService.submit(aR);
	}
	
	public void insert(Map<String, Object> aKV) {
		mExecutorService.submit(() -> {
			try {
				declareNewColumns(aKV);
				
				String lSQL = "INSERT INTO OBJECTS ("+String.join(", ", aKV.keySet().stream().map(aS -> "\""+aS+"\"").toList())+") VALUES ("+String.join(", ", Collections.nCopies(aKV.size(), "?"))+")";

				PreparedStatement lPS = mConnection.prepareStatement(lSQL, new String[]{ sJID });
				List<Object> lCol = aKV.values().stream().toList();
				for (int i = 0; i < lCol.size(); i++) lPS.setObject(i + 1, lCol.get(i));
				lPS.execute();
				
				L.d("SQL: {} with {}", lSQL, lCol);

				ResultSet lRS = lPS.getGeneratedKeys();
				if (lRS.next()) aKV.put(sJID, lRS.getInt(1));
			} catch(Exception aE) { L.w(aE); }
		});
	}
	
	public void update(Map<String, Object> aKV) {
		mExecutorService.submit(() -> {
			try {
				declareNewColumns(aKV);

				int lJID = (int)aKV.remove(sJID);
				String lSQL = "UPDATE OBJECTS SET "+String.join(", ", aKV.keySet().stream().map(aS -> "\""+aS+"\"=?").toList())+" WHERE "+sJID+"="+lJID;
				
				PreparedStatement lPS = mConnection.prepareStatement(lSQL);
				List<Object> lCol = aKV.values().stream().toList();
				for (int i = 0; i < lCol.size(); i++) lPS.setObject(i + 1, lCol.get(i));
				lPS.execute();
				
				L.d("SQL: {} with {}", lSQL, lCol);
				
				aKV.put(sJID, lJID);
			} catch(Exception aE) { L.w(aE); }
		});
	}
	
	public Stream<Map<String, Object>> query(String aS, Object ... aParams) {
		return StreamSupport.stream(Spliterators.spliteratorUnknownSize(new Iterator<Map<String, Object>>() {
			ResultSet	mResultSet;
			String		mCols[];
			
			{
				try { mResultSet = _query(aS, aParams).get(); } catch (Exception aE) { L.w(aE); }
			}
			
			@Override
			public boolean hasNext() {
				try { return !mResultSet.isAfterLast() && !mResultSet.isLast(); } catch (Exception aE) {};
				return false;
			}
			
			@Override
			public Map<String, Object> next() {
				try {
					mResultSet.next();
					
					if (mCols == null) {
						ResultSetMetaData lRSM = mResultSet.getMetaData();
						mCols = new String[lRSM.getColumnCount()];
						for (int i = 0; i < mCols.length; i++) mCols[i] = lRSM.getColumnName(i + 1);
					}
					
					Map<String, Object> lRes = new HashMap<>();
					for (int i = 0; i < mCols.length; i++) lRes.put(mCols[i], mResultSet.getObject(i + 1));

					return lRes;
				} catch (Exception aE) { L.w(aE); }
				return null;
			}
		}, Spliterator.SORTED), false);
	}
	
	protected Future<ResultSet> _query(String aS, Object ... aParams) {
		return mExecutorService.submit(() -> {
			try {
				PreparedStatement lPS = mConnection.prepareStatement(aS, ResultSet.TYPE_SCROLL_INSENSITIVE, ResultSet.CONCUR_READ_ONLY);
				if (aParams != null) for (int i=0; i<aParams.length; i++) lPS.setObject(i+1, aParams[i]);
				
				ResultSet lRS = lPS.executeQuery();
				lRS.last();
				L.t("SQL Query: {} with {} => {} lines", aS, aParams, lRS.getRow());
				lRS.beforeFirst();
				
				return lRS;

//				return lPS.executeQuery();
			} catch (Exception aE) { L.w(aE); }
			return null;
		});
	}
	
	protected void declareNewColumns(Map<String, Object> aKV) {
		Set<String> lNewCols = new HashSet<>(aKV.keySet());
		lNewCols.removeAll(mColumns);
		
		if (!lNewCols.isEmpty()) {
			try {
				String lSQL = "ALTER TABLE OBJECTS ADD COLUMN (" + String.join(", ", lNewCols.stream().map(aCol -> {
					String lType = mColumnSQLTypeResolver.apply(aCol);
					mColumns.add(aCol);
					return "\"" + aCol + "\" " + lType;
				}).toList()) + ")";
				
				mConnection.createStatement().execute(lSQL);
				L.d("SQL: " + lSQL);
			} catch (Exception aE) {
				L.w(aE);
			}
		}
	}
	
	protected Connection				mConnection;
	protected ExecutorService			mExecutorService;
	protected Set<String>				mColumns = new HashSet<>();
	protected Function<String, String>	mColumnSQLTypeResolver;
}