package fr.doap.jdb;

import fr.doap.slog.*;

import javax.sql.rowset.serial.*;
import javax.swing.plaf.*;
import java.lang.reflect.*;
import java.sql.*;
import java.util.*;
import java.util.concurrent.*;

import static fr.doap.jdb.Utils.*;

public class JDB {
	/* public static */
	static public void init(String aConn, String aUser, String aPass) {
		sJDB = new JDB();
		sJDB._init(aConn, aUser, aPass);
	}
	
	static public void save(Object aObject) {
		sJDB._save(aObject);
	}
	
	static protected JDB sJDB;
	
	/* protected member */
	protected void _init(String aConn, String aUser, String aPass) {
		mDBWrap = new DBWrap(aConn, aUser, aPass).setSQLTypeResolver(aCol -> mSQLTypes.get(aCol));
		mSQLTypes.put("_CLASS_", "VARCHAR");
		mSQLTypes.put("_SUBS_", "VARBINARY");
		
		Optional<Map<String, Object>> lOpt = mDBWrap.query("SELECT MAX("+DBWrap.sJID+")+1 AS NEXTID FROM OBJECTS").findFirst();
		Object lNID = lOpt.get().get("NEXTID");
		if (lNID != null) mNextID = (int)lNID;
		else mNextID = 1;
	}
	
	protected void _save(Object aObject) {
		new ObjectMapBuilder(aObject);
	}
	
	// aConds = String: le SQL et ensuite les paramètres
	protected <T> Iterator<T> _query(Class<T> aClass, Object ... aConds) {
		return new Iterator<T>() {
			ResultSet mRS;
			{
				try {
					mRS = _buildQuery(Utils.getShortClassname(aClass), aConds).get();
				}
				catch (Exception aE) { L.w(aE); }
			}
			
			@Override
			public boolean hasNext() {
				try { return !(mRS.isLast() || mRS.isAfterLast()); }
				catch (Exception aE) { return false; }
			}
			
			@Override
			public T next() {
				try {
					mRS.next();
					int lID = mRS.getInt(mDBWrap.sJID);
L.i("ID: {}", lID);
					Object lCached = mWeakObjectCache.getObject(lID);
					if (lCached != null) return (T)lCached;
L.i("NOT CACHED ! Loading...");
					
					Map<Integer, Map<String, Object>> lProps = new HashMap<>();
					Set<String> lCols = new HashSet<>();
					int lColCount = mRS.getMetaData().getColumnCount();
					for (int i=0; i<lColCount; i++) lCols.add(mRS.getMetaData().getColumnName(i+1));
L.i("Cols: {}", lCols);
					
					Map<String, Object> lMap = new HashMap<>();
					lProps.put(lID, lMap);
					for (String iCol: lCols) {
						Object lP = mRS.getObject(iCol);
						if (lP != null) lMap.put(iCol, lP);
					}
L.i("Props: {}", lMap);
					
					List<Integer> lLoadList = new ArrayList<>();
					byte[] lBA = mRS.getBytes("_SUBS_");
					if (lBA != null) {
						int[] lIDs = (int[]) Utils.fromByteArray(lBA);
						for (int i=0; i<lIDs.length; i++) if (mWeakObjectCache.getObject(lIDs[i]) == null) lLoadList.add(lIDs[i]);
					}
					
					ResultSet lLoadRS = mDBWrap._query("SELECT * FROM OBJECTS WHERE \""+DBWrap.sJID+"\" IN ("+String.join(",", lLoadList.stream().map(aID -> aID.toString()).toList())+")").get();
					while (!lLoadRS.isLast() && !lLoadRS.isAfterLast()) {
						lLoadRS.next();
						
						lMap = new HashMap<>();
						int lSID = lLoadRS.getInt(mDBWrap.sJID);
						lProps.put(lSID, lMap);
						for (String iCol: lCols) {
							Object lP = lLoadRS.getObject(iCol);
							if (lP != null) lMap.put(iCol, lP);
						}
L.i(" --> Props: {}", lMap);
						
						Class<?> lClass = Class.forName(lLoadRS.getString("_CLASS_"));
						Object lO = lClass.getDeclaredConstructor().newInstance();
						mWeakObjectCache.put(lSID, lO);
					}
					lProps.forEach((aID, aProps) -> {
						Object lO = mWeakObjectCache.getId(aID);
						if ((lO instanceof Collection)&&(lO.getClass().getCanonicalName().startsWith("java."))) {
						
						} else {
							aProps.forEach((aK, aV) -> {
								if (!aK.startsWith("_") && aK.contains("#")) {
									String[] lKs = aK.split("#");
									try {
										Field lF = Class.forName(lKs[0]).getDeclaredField(lKs[1]);
										L.i("- key: {} -> {}", aK, lF);
									} catch (Exception aE) { L.t(aE); }
								}
							});
						}
					});
					
					return (T)mWeakObjectCache.getObject(lID);
				} catch (Exception aE) { L.w(aE); }
				return null;
			}
		};
	}
	
	protected Future<ResultSet> _buildQuery(String aClassname, Object ... aConds) {
		Object[] lParams = null;
		String lSQL = "SELECT * FROM OBJECTS WHERE \""+aClassname+"\"=TRUE";
		if ((aConds != null)&&(aConds.length > 0)) {
			lSQL += " AND "+(String)aConds[0];
			if (aConds.length > 1) {
				lParams = new Object[aConds.length - 1];
				for (int i = 0; i < lParams.length; i++) lParams[i] = aConds[i + 1];
			}
		}
		return mDBWrap._query(lSQL, lParams);
	}
	
	static final Map<Class<?>, String> sTypeMap = Utils.HashMapOf(
		Boolean.class,	"BOOLEAN",
		Integer.class,	"INTEGER",
		Long.class,		"BIGINT",
		byte[].class,	"VARBINARY",
		Blob.class,		"BLOB",
		String.class,	"VARCHAR");
	
	class ObjectMapBuilder {
		ObjectMapBuilder(Object aObject) {
			Map<Object, Set<Object>> lReach=new HashMap<>();
			Map<Object, Map<String, Object>> lInserts=new HashMap<>();
			Map<Object, Map<String, Object>> lUpdates=new HashMap<>();
			
			add(aObject);
			
			while (mPos < mProcessingList.size()) {
				Object lO = mProcessingList.get(mPos);
				try {
					Constructor lCons = lO.getClass().getDeclaredConstructor(); // ensure at least a default constructor...
				} catch (Exception aE) {
					throw new RuntimeException("No constructor for class "+lO.getClass().getCanonicalName());
				}

				Set<Object> lObjects = new HashSet<>();
				lReach.put(lO, lObjects);
				
				Map<String, Object> lProps = getProperties(lO);
				for (Map.Entry<String, Object> iME: lProps.entrySet()) {
					Object lV = iME.getValue();
					if (!sTypeMap.containsKey(lV.getClass())) {
						add(lV);
						iME.setValue(mWeakObjectCache.getId(lV));
						mSQLTypes.put(iME.getKey(), "INTEGER");
						lObjects.add(lV);
					} else mSQLTypes.put(iME.getKey(), sTypeMap.get(lV.getClass()));
				}
				
				lProps.put(DBWrap.sJID, mWeakObjectCache.getId(lO));
				lProps.put("_CLASS_", lO.getClass().getCanonicalName());

				if (mInserts.contains(lO)) lInserts.put(lO, lProps);
				else lUpdates.put(lO, lProps);
				
				mPos++;
			}
			
			// saturate
			Set<Object> lReachComplete = new HashSet<>();
			lReachComplete.add(this);
			while (!lReachComplete.isEmpty()) {
				lReachComplete.clear();
				lReach.values().forEach(aSet -> {
					Set<Object> lNew = new HashSet<>(aSet);
					aSet.forEach(aE -> {
						if (lNew.addAll(lReach.get(aE))) lReachComplete.add(aE);
					});
					aSet.clear();
					aSet.addAll(lNew);
				});
			}

			lReach.forEach((aO, aSet) -> {
				if (aSet.size() > 0) {
					Object[] lObs = aSet.toArray();
					int lNew[] = new int[lObs.length];
					for (int i = 0; i < lNew.length; i++) lNew[i] = mWeakObjectCache.getId(lObs[i]);
L.i("Reach {} -> {}", aSet, lNew);
					if (lInserts.containsKey(aO)) lInserts.get(aO).put("_SUBS_", Utils.toByteArray(lNew));
					else lUpdates.get(aO).put("_SUBS_", Utils.toByteArray(lNew));
				}
			});

			L.i("Reach: {}", lReach);
			lInserts.values().forEach(aProps -> mDBWrap.insert(aProps));
			lUpdates.values().forEach(aProps -> mDBWrap.update(aProps));
		}

		private void add(Object aObject) {
			if (!mProcessingList.contains(aObject)) {
				mProcessingList.add(aObject);
				if (mWeakObjectCache.getId(aObject) == -1) {
					mWeakObjectCache.put(mNextID, aObject);
					mNextID++;
					mInserts.add(aObject);
				}
			}
		}
		
		protected Map<String, Object> getProperties(Object aObject) {
			Map<String, Object> lProps = new HashMap<>();
			boolean isCollection = aObject instanceof Collection;
			Class<?> lC = aObject.getClass();
			while (lC != Object.class) {
				lProps.put(getShortClassname(lC), true);
				if (isCollection && lC.getCanonicalName().startsWith("java.")) {
					Collection lColl = (Collection) aObject;
					Object[] lObjects = new Object[2*lColl.size()];
					Iterator iIt = lColl.iterator();
					for (int i=0; i<lColl.size(); i++) {
						Object lV = iIt.next();
						if (!sTypeMap.containsKey(lV.getClass())) {
							add(lV);
							lObjects[2*i] = null;
							lObjects[2*i+1] = mWeakObjectCache.getId(lV);
						} else {
							lObjects[2*i] = lV.getClass().getCanonicalName();
							lObjects[2*i+1] = lV;
						}
					}
L.i("Coded collection is {}", Arrays.asList(lObjects));
					lProps.put("_CONTENT_", Utils.toByteArray(lObjects));
					break;
				} else {
					Field[] lFields = lC.getDeclaredFields();
					for (Field iF : lFields) {
						int lModif = iF.getModifiers();
						if (((lModif & Modifier.FINAL) | (lModif & Modifier.TRANSIENT) | (lModif & Modifier.VOLATILE)) == 0) {
							if (iF.trySetAccessible()) {
								try {
									lProps.put(getShortClassname(lC) + "#" + iF.getName(), iF.get(aObject));
								} catch (Exception aE) {
									L.t(aE);
								}
							}
						}
					}
					lC = lC.getSuperclass();
				}
			}
			return lProps;
		}
		
		int							mPos = 0;
		List<Object>				mProcessingList = new ArrayList<>();
		Set<Object>					mInserts = new HashSet<>();
	}
	
	protected DBWrap				mDBWrap;
	protected WeakObjectCache		mWeakObjectCache = new WeakObjectCache();
	protected int					mNextID;
	protected Map<String, String>	mSQLTypes = new HashMap<>();
}